from typing import TYPE_CHECKING, Callable, MutableMapping

if TYPE_CHECKING:
    from .state import State

from nbterm import Notebook

NotebookExecutor = Callable[["State"], None]

NOTEBOOKS: MutableMapping[str, NotebookExecutor] = {}


def notebook(name: str) -> Callable[[NotebookExecutor], NotebookExecutor]:
    def registrar(executor: NotebookExecutor) -> NotebookExecutor:
        NOTEBOOKS[name] = executor
        return executor

    return registrar


@notebook("nbterm")
def nbterm(state: "State") -> None:
    save_path = state.tmp_dir / f"{id(state.current_slide)}.ipynb"

    nb = Notebook(state.current_slide.notebook or save_path)

    state.current_slide.notebook = save_path

    nb.show()
    nb.save(save_path)



import inspect
import os
import shutil
import socket
from datetime import datetime
from math import cos, floor, pi
from pathlib import Path
from textwrap import dedent

import numpy as np
from rich.align import Align
from rich.box import SQUARE
from rich.color import Color, blend_rgb
from rich.console import RenderGroup
from rich.layout import Layout
from rich.markdown import Markdown
from rich.padding import Padding
from rich.panel import Panel
from rich.style import Style
from rich.syntax import Syntax
from rich.text import Text

from spiel import Deck, Image, Options, Slide, __version__, example_panels
from spiel.plot import Plot

deck = Deck(name=f"Spiel Demo Deck (v{__version__})")
options = Options()

SPIEL = "[Spiel](https://github.com/JoshKarpel/spiel)"
RICH = "[Rich](https://rich.readthedocs.io/)"
NBTERM = "[nbterm](https://github.com/davidbrochart/nbterm)"
UNIPLOT = "[Uniplot](https://github.com/olavolav/uniplot)"
IPYTHON = "[IPython](https://ipython.readthedocs.io)"
WSL = "[Windows Subsystem for Linux](https://docs.microsoft.com/en-us/windows/wsl/)"

THIS_DIR = Path(__file__).resolve().parent


@deck.slide(title="What is Spiel?")
def what():
    upper_left_markup = dedent(
        f"""\
        ## What is Spiel?

        {SPIEL} is a framework for building and presenting richly-styled presentations in your terminal using Python.

        Spiel uses {RICH} to render slide content.
        Anything you can display with Rich, you can display with Spiel (plus some other things)!

        Use your right `→` and left `←` arrows keys (or `f` and `b`) to go forwards and backwards through the deck.

        Press `ctrl-c` or `ctrl-k` to exit.

        Press `h` at any time to see the help screen, which describes all of the actions you can take.

        To get a copy of the source code for this deck, use the `spiel demo copy` command.
        """
    )

    upper_right_markup = dedent(
        """\
        ## Why use Spiel?

        It's fun!

        It's weird!

        Why not?

        Maybe you shouldn't.

        Honestly, it's unclear whether it's a good idea.

        There's always [Powerpoint](https://youtu.be/uNjxe8ShM-8)!
        """
    )

    lower_left_markup = dedent(
        f"""\
        ## Contributing

        Please report bugs via [GitHub Issues](https://github.com/JoshKarpel/spiel/issues).

        If you have ideas about how Spiel can be improved,
        or you have a cool deck to show off,
        please post to [GitHub Discussions](https://github.com/JoshKarpel/spiel/discussions).
        """
    )

    lower_right_markup = dedent(
        f"""\
        ## Inspirations

        Brandon Rhodes' [PyCon 2017](https://youtu.be/66P5FMkWoVU) and [North Bay Python 2017](https://youtu.be/rrMnmLyYjU8) talks.

        David Beazley's [Lambda Calculus from the Ground Up](https://youtu.be/pkCLMl0e_0k) tutorial at PyCon 2019.

        LaTeX's [Beamer](https://ctan.org/pkg/beamer) document class.
        """
    )

    def pad_markdown(markup):
        return Padding(Markdown(markup, justify="center"), pad=(0, 5))

    root = Layout()
    upper = Layout()
    lower = Layout()

    upper.split_row(
        Layout(pad_markdown(upper_left_markup)),
        Layout(pad_markdown(upper_right_markup)),
    )
    lower.split_row(
        Layout(pad_markdown(lower_left_markup)),
        Layout(pad_markdown(lower_right_markup)),
    )
    root.split_column(upper, lower)

    return root


@deck.slide(title="Decks and Slides")
def code():
    markup = dedent(
        f"""\
    ## Decks are made of Slides

    Here's the code for `Deck` and `Slide`!

    The source code is pulled directly from the definitions via [`inspect.getsource`](https://docs.python.org/3/library/inspect.html#inspect.getsource).

    ({RICH} supports syntax highlighting, so {SPIEL} does too!)
    """
    )
    root = Layout()
    upper = Layout(Markdown(markup, justify="center"), size=len(markup.split("\n")) + 1)
    lower = Layout()
    root.split_column(upper, lower)

    def make_code_panel(obj):
        lines, line_number = inspect.getsourcelines(obj)
        return Panel(
            Syntax(
                "".join(lines),
                lexer_name="python",
                line_numbers=True,
                start_line=line_number,
            ),
            box=SQUARE,
            border_style=Style(dim=True),
            height=len(lines) + 2,
        )

    lower.split_row(
        Layout(make_code_panel(Deck)),
        Layout(make_code_panel(Slide)),
    )

    return root


@deck.slide(title="Dynamic Content")
def dynamic():
    home = Path.home()
    width = shutil.get_terminal_size().columns
    width_limit = 80
    return RenderGroup(
        Align(
            Text(
                f"Slides can have dynamic content!",
                style=Style(color="bright_magenta", bold=True, italic=True),
                justify="center",
            ),
            align="center",
        ),
        Align(
            Panel(
                Text(
                    f"Your terminal is {width} cells wide."
                    if width > width_limit
                    else f"Your terminal is only {width} cells wide! Get a bigger monitor!",
                    style=Style(color="green1" if width > width_limit else "red"),
                    justify="center",
                )
            ),
            align="center",
        ),
        Align(
            Panel(
                Text(
                    f"The time on this computer, {socket.gethostname()}, is {datetime.now()}",
                    style=Style(color="bright_cyan", bold=True, italic=True),
                    justify="center",
                )
            ),
            align="center",
        ),
        Align(
            Panel(
                Text(
                    f"There are {len([f for f in home.iterdir() if f.is_file()])} files in {home} right now.",
                    style=Style(color="yellow"),
                    justify="center",
                )
            ),
            align="center",
        ),
    )


@deck.slide(title="Triggers")
def triggers(triggers):
    info = Markdown(
        dedent(
            f"""\
            ## Triggers

            Triggers are a mechanism for making dynamic content that depends on *relative* time.

            Triggers can be used to implement effects like fades, motion, and other "animations".

            Each slide is triggered once when it starts being displayed.
            You can trigger it again (as many times as you'd like) by pressing `t`.
            You can reset the trigger state by pressing `r`.

            This slide has been triggered {len(triggers)} times.
            It was last triggered {triggers.time_since_last_trigger:.2f} seconds ago.
            """
        ),
        justify="center",
    )

    bounce_period = 10
    width = 50
    half_width = width // 2

    bounce_time = triggers.time_since_first_trigger % bounce_period
    bounce_character = "⁍" if bounce_time < (1 / 2) * bounce_period else "⁌"
    bounce_position = floor(half_width * cos(2 * pi * bounce_time / bounce_period))
    before = half_width + bounce_position
    ball = Align.center(
        Panel(
            Padding(
                bounce_character,
                pad=(0, before, 0, (half_width - bounce_position - 1)),
            ),
            title="Bouncing Bullet",
            padding=0,
        )
    )

    white = Color.parse("bright_white")
    black = Color.parse("black")
    red = Color.parse("bright_red")
    green = Color.parse("bright_green")

    fade_time = 3

    lines = [
        Text(
            "Triggered!",
            style=Style(
                color=(
                    Color.from_triplet(
                        blend_rgb(
                            black.get_truecolor(),
                            white.get_truecolor(),
                            cross_fade=min((triggers.now - time) / fade_time, 1),
                        )
                    )
                )
            ),
        )
        for time in triggers.times
    ]

    fun = Align.center(
        Panel(
            Text("\n", justify="center").join(lines),
            border_style=Style(
                color=Color.from_triplet(
                    blend_rgb(
                        green.get_truecolor(),
                        red.get_truecolor(),
                        cross_fade=min(triggers.time_since_last_trigger / fade_time, 1),
                    )
                ),
            ),
            title="Trigger Tracker",
        )
    )
    return RenderGroup(info, fun, ball if len(triggers) > 2 else Text(""))


@deck.slide(title="Views")
def grid():
    markup = dedent(
        """\
    ## Multiple Views

    Try pressing `d` to go into "deck" view.
    Press `s` to go back to "slide" view.

    Press `j`, then enter a slide number (like `3`) to jump to a slide.
    """
    )
    return Markdown(markup, justify="center")


@deck.slide(title="Watch Mode")
def watch():
    markup = dedent(
        f"""\
    ## Developing a Deck

    {SPIEL} can reload your deck as you edit it if you add the `--watch` option to `present`:

    `$ spiel present path/to/deck.py --watch`

    If you're on a system without inotify support (e.g., {WSL}), you should use the `--poll` option instead.
    """
    )
    return Markdown(markup, justify="center")


@deck.slide(title="Displaying Images")
def image():
    markup = dedent(
        f"""\
    ## Images

    {SPIEL} can display images... sort of!

    Spiel includes an `Image` widget that can render images by interpolating pixel values.

    If you see big chunks of constant color instead of smooth gradients, your terminal is probably not configured for "truecolor" mode.
    If your terminal supports truecolor (it probably does), try setting the environment variable `COLORTERM` to `truecolor`.
    For example, for `bash`, you could add

    `export COLORTERM=truecolor`

    to your `.bashrc` file, then restart your shell.
    """
    )
    root = Layout()
    root.split_row(
        Layout(Padding(Markdown(markup, justify="center"), pad=(0, 2))),
        Layout(Image.from_file(THIS_DIR / "img.jpg")),
    )

    return root


@deck.slide(title="Plots")
def plots(triggers):
    markup = dedent(
        f"""\
        ## Plots

        {SPIEL} can display plots... sort of!

        Spiel includes a `Plot` widget that uses {UNIPLOT} to render plots.

        You can even make animated plots by using triggers! Try triggering this slide.
        """
    )

    x = np.linspace(-3, 3, 1000)
    coefficients = np.array([0, 2, -1, 3, -1, 1])
    dither = 0.1 * np.sin(triggers.time_since_last_trigger) if triggers.triggered else 0
    hermite = 0.9 * np.polynomial.hermite.hermval(x, c=coefficients + dither)
    upper_plot = Plot(xs=x, ys=hermite, title="A Hermite Polynomial", y_min=-100, y_max=100)

    theta = np.linspace(-3 * np.pi, 3 * np.pi, 1000)
    phase = (1 if triggers.triggered else 0) * triggers.time_since_last_trigger
    cos = np.cos(theta + phase)
    sin = 1.3 * np.sin(theta + phase)
    lower_plot = Plot(xs=[theta, theta], ys=[cos, sin], title="[cos(x), 1.3 * sin(x)]")

    lower = Layout()
    lower.split_column(
        Layout(upper_plot),
        Layout(lower_plot),
    )

    root = Layout()
    root.split_row(
        Layout(Padding(Markdown(markup, justify="center"), pad=(0, 2))),
        lower,
    )

    return root


@deck.example(title="Examples")
def examples():
    # This is an example that shows how to use random.choice from the standard library.

    # The source code is embedded directly into the demo deck file,
    # but you could load it from another file if you wanted to.

    import random

    directions = ["North", "South", "East", "West"]

    print("Which way should we go?")
    print(random.choice(directions))


@examples.layout
def _(example, triggers):
    root = Layout()

    extra = (
        f"""
    ## Example Execution is Cached

    Now that you've triggered the slide, {SPIEL} will execute the example once and display the output.
    The result is cached, so the example is not executed on every frame, like code in normal slide content
    functions is.

    ## Editing Examples

    Examples can be modified during the talk.
    Press `e` to open your `$EDITOR` (`{os.getenv("EDITOR", "not set")}`) on the example code.
    Save your changes and exit to come back to the presentation with your updated code.
    You can then trigger the example again to run it with the new code.

    ## Layout Customization

    You can customize the example slide's content by providing a custom `layout` function.
    If you don't, you'll get the default layout, which looks like just the right half of this slide.
    """
        if triggers.triggered
        else ""
    )

    markup = dedent(
        f"""\
    ## Examples

    {SPIEL} can display and execute chunks of example code.

    Example slides are driven by the trigger system.
    Press `t` to execute the example code and display the output.

    {extra}
    """
    )
    markdown = Markdown(markup, justify="center")

    root.split_row(
        Layout(Padding(markdown, pad=(0, 2))),
        example_panels(example),
    )

    return root


@deck.slide(title="Live Coding with the REPL")
def repl():
    markup = dedent(
        f"""\
        ## Live Coding: REPL

        Sometimes an static example,
        or even an example that you're editing and running multiple times,
        just isn't interactive enough.

        To provide a more interactive experience,
        {SPIEL} lets you open a REPL on any slide by pressing `i`.

        There are two REPLs available by default: the [builtin Python REPL](https://docs.python.org/3/tutorial/interpreter.html#interactive-mode) and {IPYTHON}.
        You can change which REPL to use via `Options`, which will be discussed later.

        When you exit the REPL (by pressing `ctrl-d` or executing `exit`),
        you'll be back at the same point in your presentation.

        The state of the REPL is not persistent between invocations
        (it will be completely fresh every time you enter it).
        """
    )
    return Markdown(markup, justify="center")


@deck.slide(title="Live Coding with Notebooks", notebook=THIS_DIR / "notebook.ipynb")
def notebooks():
    markup = dedent(
        f"""\
        ## Live Coding: Notebooks

        For a more persistent live-coding experience than a REPL,
        you can open a Jupyter Notebook via {NBTERM}
        by pressing `n`.

        Each slide has a notebook attached to it.
        The notebook kernel will be restarted between invocations,
        but changes made to the notebook contents
        will persist throughout your presentation.

        By default, the notebook will initially be blank,
        but you can also provide a path to a notebook on disk to initialize from.

        This slide has a small example notebook attached to it - try it out!

        (In theory, {SPIEL} will be able to support different kinds of terminal notebook viewers in the future,
        but right now, only {NBTERM} is packaged by default, and the internal API is not stable or very generic.
        Stay tuned!)
        """
    )
    return Markdown(markup, justify="center")


@deck.slide(title="Options")
def options_():
    markup = dedent(
        f"""\
        ## Options

        {SPIEL} has a variety of options that can be adjusted at runtime.
        For example,
        profiling information can be displayed in the footer to help you debug a slides that is rendering too slowly.

        To see your current options, press `p`.
        From that mode you can edit your options by pressing `e`.

        Note that your `Options` are *not* reloaded when running with `--watch`.
        """
    )
    return Markdown(markup, justify="center")



import shutil
from contextlib import nullcontext
from pathlib import Path
from textwrap import dedent

from rich.console import Console
from rich.control import Control
from rich.style import Style
from rich.syntax import Syntax
from rich.text import Text
from typer import Argument, Exit, Option, Typer

from .constants import PACKAGE_NAME, __version__
from .help import version_details
from .load import DeckWatcher
from .modes import Mode
from .present import present_deck
from .reloader import DeckReloader
from .state import State

THIS_DIR = Path(__file__).resolve().parent

app = Typer(
    help=dedent(
        f"""\
        Display richly-styled presentations using your terminal.

        To see what {PACKAGE_NAME.capitalize()} can do, take a look at the demo deck:

            $ spiel demo present

        A {PACKAGE_NAME.capitalize()} presentation (a "deck [of slides]") is defined programmatically using a Python script.
        """
    )
)


@app.command()
def present(
    path: Path = Argument(
        ...,
        dir_okay=False,
        help="The path to the slide deck file.",
    ),
    mode: Mode = Option(
        default=Mode.SLIDE,
        help="The mode to start presenting in.",
    ),
    slide: int = Option(
        default=1,
        help="The slide number to start the presentation on.",
    ),
    watch: bool = Option(
        default=False,
        help="If enabled, reload the deck when the slide deck file changes.",
    ),
    poll: bool = Option(
        default=False,
        help="If enabled, poll the filesystem for changes (implies --watch). Use this option on systems that don't support file modification notifications.",
    ),
) -> None:
    """
    Present a deck.
    """
    _present(path=path, mode=mode, slide=slide, watch=watch, poll=poll)


def _present(path: Path, mode: Mode, slide: int, watch: bool, poll: bool) -> None:
    console = Console()

    try:
        state = State.from_file(path)
    except FileNotFoundError as e:
        console.print(Text(f"Error: {e}", style=Style(color="red")))
        raise Exit(code=1)

    state.mode = mode
    state.jump_to_slide(slide - 1)

    watcher = (
        DeckWatcher(event_handler=DeckReloader(state, path), path=path, poll=poll)
        if (watch or poll)
        else nullcontext()
    )

    try:
        with state, watcher:
            present_deck(state)
    except KeyboardInterrupt:
        raise Exit(code=0)
    finally:
        state.console.print(Control.clear())
        state.console.print(Control.move_to(0, 0))


@app.command()
def init(
    path: Path = Argument(
        ...,
        writable=True,
        resolve_path=True,
        help="The path to create a new deck script at.",
    )
) -> None:
    """
    Create a new deck script at the given path from a basic template.

    This is a good starting point if you already know what you want to do.
    If you're not so sure, consider taking a look at the demo deck to see what's possible:

        $ spiel demo --help
    """
    console = Console()

    if path.exists():
        console.print(
            Text(f"Error: {path} already exists, refusing to overwrite.", style=Style(color="red"))
        )
        raise Exit(code=1)

    name = path.stem.replace("_", " ").title()

    try:
        path.parent.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        console.print(
            Text(
                f"Error: was not able to ensure that the parent directory {path.parent} exists due to: {e}.",
                style=Style(color="red"),
            )
        )
        raise Exit(code=1)

    try:
        path.write_text(
            dedent(
                f"""\
                from textwrap import dedent

                from spiel import Deck, Options

                deck = Deck(name="{name}")
                options = Options()

                @deck.slide(title="Title")
                def title():
                    markup = dedent(
                        \"""\\
                        # {name}

                        This is your title slide!
                        \"""
                    )

                    return Markdown(markup, justify="center")
                """
            )
        )
    except Exception as e:
        console.print(
            Text(
                f"Error: was not able to write template to {path} due to: {e}",
                style=Style(color="red"),
            )
        )
        raise Exit(code=1)

    console.print(Text(f"Wrote deck template to {path}", style=Style(color="green")))


@app.command()
def version(
    plain: bool = Option(
        default=False,
        help=f"Print only {PACKAGE_NAME}'s version.",
    )
) -> None:
    """
    Display version and debugging information.
    """
    console = Console()

    if plain:
        print(__version__)
    else:
        console.print(version_details(console))


demo = Typer(
    name="demo",
    help=dedent(
        """\
        Use the demonstration deck (present it, display source, etc.).
        """
    ),
)

DEMO_DIR = THIS_DIR / "demo"
DEMO_SOURCE = THIS_DIR / "demo" / "demo.py"


@demo.command(name="present")
def present_demo() -> None:
    """
    Present the demo deck.
    """
    _present(path=DEMO_SOURCE, mode=Mode.SLIDE, slide=0, watch=False, poll=False)


@demo.command()
def source() -> None:
    """
    Display the source code for the demo deck in your PAGER.
    """
    console = Console()

    with console.pager(styles=True):
        console.print(Syntax(DEMO_SOURCE.read_text(), lexer_name="python"))


@demo.command()
def copy(
    path: Path = Argument(
        default=...,
        writable=True,
        help="The path to copy the demo deck source code and assets to.",
    )
) -> None:
    """
    Copy the demo deck source code and assets to a new directory.

    If you're looking for a more stripped-down starting point, try the init command:

        $ spiel init --help
    """
    console = Console()

    if path.exists():
        console.print(Text(f"Error: {path} already exists!", style=Style(color="red")))
        raise Exit(code=2)

    try:
        shutil.copytree(DEMO_DIR, path)
    except Exception as e:
        console.print(Text(f"Failed to copy demo deck directory: {e}", style=Style(color="red")))
        raise Exit(code=1)

    console.print(
        Text(f"Wrote demo deck source code and assets to {path}", style=Style(color="green"))
    )


app.add_typer(demo)



from __future__ import annotations

import inspect
import sys
from collections.abc import Collection
from dataclasses import dataclass, field
from pathlib import Path
from textwrap import dedent
from typing import Callable, Iterator, List, Optional, Sequence

from .example import Example
from .presentable import Presentable
from .slide import MakeRenderable, Slide


@dataclass
class Deck(Collection):
    name: str
    slides: List[Presentable] = field(default_factory=list)

    def __getitem__(self, idx: int) -> Presentable:
        return self.slides[idx]

    def __len__(self) -> int:
        return len(self.slides)

    def __iter__(self) -> Iterator[Presentable]:
        return iter(self.slides)

    def __contains__(self, obj: object) -> bool:
        return obj in self.slides

    def add_slides(self, *slides: Presentable) -> Deck:
        self.slides.extend(slides)
        return self

    def slide(
        self,
        title: str = "",
        notebook: Optional[Path] = None,
    ) -> Callable[[MakeRenderable], Slide]:
        def slideify(content: MakeRenderable) -> Slide:
            slide = Slide(
                title=title,
                notebook=notebook,
                content=content,
            )
            self.add_slides(slide)
            return slide

        return slideify

    def example(
        self,
        title: str = "",
        command: Sequence[str] = (sys.executable,),
        name: str = "example.py",
        language: str = "python",
        notebook: Optional[Path] = None,
    ) -> Callable[[Callable], Example]:
        def exampleify(example: Callable) -> Example:
            ex = Example(
                title=title,
                notebook=notebook,
                source=get_function_body(example),
                command=command,
                name=name,
                language=language,
            )
            self.add_slides(ex)
            return ex

        return exampleify


def get_function_body(function: Callable) -> str:
    lines, _ = inspect.getsourcelines(function)

    prev_indent = None
    for idx, line in enumerate(lines):
        if prev_indent is None:
            prev_indent = count_leading_whitespace(line)
        elif count_leading_whitespace(line) > prev_indent:
            return dedent("".join(lines[idx:]))

    raise ValueError(f"Could not extract function body from {function}")


def count_leading_whitespace(s: str) -> int:
    return len(s) - len(s.lstrip())



from __future__ import annotations

import importlib.util
import sys
from dataclasses import dataclass
from pathlib import Path
from types import TracebackType
from typing import ContextManager, Optional, Tuple, Type

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from watchdog.observers.polling import PollingObserver

from .constants import DECK, OPTIONS
from .deck import Deck
from .exceptions import NoDeckFound
from .options import Options


def load_deck_and_options(path: Path) -> Tuple[Deck, Options]:
    module_name = "__deck"
    spec = importlib.util.spec_from_file_location(module_name, path)

    if spec is None:
        raise FileNotFoundError(
            f"{path.resolve()} does not appear to be an importable Python module."
        )

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)  # type: ignore

    try:
        deck = getattr(module, DECK)
    except AttributeError:
        raise NoDeckFound(f"The module at {path} does not have an attribute named {DECK}.")

    if not isinstance(deck, Deck):
        raise NoDeckFound(
            f"The module at {path} has an attribute named {DECK}, but it is a {type(deck).__name__}, not a {Deck.__name__}."
        )

    options = getattr(module, OPTIONS, Options())

    if not isinstance(options, Options):
        options = Options()

    return deck, options


@dataclass
class DeckWatcher(ContextManager):
    event_handler: FileSystemEventHandler
    path: Path
    poll: bool = False

    def __post_init__(self) -> None:
        self.observer = (PollingObserver if self.poll else Observer)(timeout=0.1)

    def __enter__(self) -> DeckWatcher:
        self.observer.schedule(self.event_handler, str(self.path))
        self.observer.start()

        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        self.observer.stop()
        self.observer.join()

        return None



from .constants import __version__
from .deck import Deck
from .example import Example, example_panels
from .image import Image
from .notebooks import notebook
from .options import Options
from .repls import repl
from .slide import Slide
from .triggers import Triggers



from itertools import zip_longest
from typing import Any, Iterable, Iterator, Optional, TypeVar

T = TypeVar("T")


def joinify(joiner: str, items: Iterable[Optional[Any]]) -> str:
    return joiner.join(map(str, filter(None, items)))


def drop_nones(*items: Optional[T]) -> Iterator[T]:
    yield from (item for item in items if item is not None)


def clamp(value: int, lower: int, upper: int) -> int:
    if lower > upper:
        raise ValueError(
            f"Upper bound ({upper}) for clamp must be greater than lower bound ({lower})."
        )
    return max(min(value, upper), lower)


def chunks(iterable: Iterable[T], n: int, fill_value: Optional[T] = None) -> Iterable[Iterable[T]]:
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fill_value)



import os
import sys
from importlib import metadata

PACKAGE_NAME = "spiel"
__version__ = metadata.version(PACKAGE_NAME)
__rich_version__ = metadata.version("rich")
__python_version__ = ".".join(map(str, sys.version_info))

DECK = "deck"
OPTIONS = "options"

TARGET_RPS = 30

EDITOR = os.getenv("EDITOR", "not set")



import re
from typing import Any, Iterable, Sequence, Union

import numpy as np
import uniplot
from colorama import Fore
from colorama import Style as CStyle
from rich.console import Console, ConsoleOptions
from rich.segment import Segment
from rich.style import Style
from rich.text import Text

Plottable = Union[
    np.ndarray,
    Sequence[np.ndarray],
]

RE_ANSI_ESCAPE = re.compile(r"(\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~]))")

COLOR_MAP = {
    Fore.RED: "red",
    Fore.GREEN: "green",
    Fore.BLUE: "blue",
    Fore.CYAN: "cyan",
    Fore.YELLOW: "yellow",
    Fore.MAGENTA: "magenta",
    Fore.BLACK: "black",
    Fore.WHITE: "white",
}


class Plot:
    def __init__(
        self,
        **plot_args: Any,
    ) -> None:
        self.plot_args = plot_args

    def _ansi_to_text(self, s: str) -> Text:
        pieces = []
        tmp = ""
        style = Style.null()
        for char in RE_ANSI_ESCAPE.split(s):
            if char == CStyle.RESET_ALL:
                pieces.append(Text(tmp, style=style))
                style = Style.null()
                tmp = ""
            elif char in COLOR_MAP:
                pieces.append(Text(tmp, style=style))
                style = Style(color=COLOR_MAP[char])
                tmp = ""
            else:
                tmp += char

        # catch leftovers
        pieces.append(Text(tmp, style=style))

        return Text("", no_wrap=True).join(pieces)

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> Iterable[Segment]:
        if self.plot_args.get("height") is None and options.height is None:
            height = None
        else:
            # 5 = title + top bar + bottom bar + bottom axis labels + 1
            height = max(
                (options.height - 5) if options.height else 1, self.plot_args.get("height", 1)
            )

        plot_args = {
            **self.plot_args,
            **dict(
                height=height,
                width=max(options.max_width - 10, self.plot_args.get("width", 1)),
            ),
        }

        plot = "\n".join(uniplot.plot_to_string(**plot_args))

        text = self._ansi_to_text(plot)

        yield from text.__rich_console__(console, options)



import code
from typing import Callable, MutableMapping

import IPython
from traitlets.config import Config

REPLExecutor = Callable[[], None]

REPLS: MutableMapping[str, REPLExecutor] = {}


def repl(name: str) -> Callable[[REPLExecutor], REPLExecutor]:
    def registrar(executor: REPLExecutor) -> REPLExecutor:
        REPLS[name] = executor
        return executor

    return registrar


@repl("builtin")
def builtin() -> None:
    code.InteractiveConsole().interact()


@repl("ipython")
def ipython() -> None:
    c = Config()
    c.InteractiveShellEmbed.colors = "Neutral"

    IPython.embed(config=c)



from dataclasses import dataclass, field
from functools import cached_property
from time import monotonic
from typing import Iterator, Tuple


@dataclass(frozen=True)
class Triggers:
    times: Tuple[float, ...]
    now: float = field(default_factory=monotonic)

    def __len__(self) -> int:
        return len(self.times)

    def __getitem__(self, idx: int) -> float:
        return self.times[idx]

    def __iter__(self) -> Iterator[float]:
        return iter(self.times)

    @cached_property
    def time_since_last_trigger(self) -> float:
        return self.now - self.times[-1]

    @cached_property
    def time_since_first_trigger(self) -> float:
        return self.now - self.times[0]

    @cached_property
    def triggered(self) -> bool:
        return len(self) > 1



import sys
from itertools import islice
from math import ceil
from time import monotonic

from rich.console import ConsoleRenderable
from rich.layout import Layout
from rich.live import Live
from rich.padding import Padding
from rich.panel import Panel
from rich.style import Style

from .constants import TARGET_RPS
from .exceptions import UnknownModeError
from .footer import Footer
from .help import Help
from .input import handle_input, no_echo
from .modes import Mode
from .presentable import Presentable
from .rps import RPSCounter
from .state import State
from .triggers import Triggers
from .utils import clamp, joinify


def render_slide(state: State, slide: Presentable) -> ConsoleRenderable:
    return Padding(
        slide.render(triggers=Triggers(times=tuple(state.trigger_times))),
        pad=1,
    )


def split_layout_into_deck_grid(root: Layout, state: State) -> Layout:
    grid_width = state.deck_grid_width
    row_of_current_slide = state.current_slide_idx // grid_width
    num_rows = ceil(len(state.deck) / grid_width)
    start_row = clamp(
        value=row_of_current_slide - (grid_width // 2),
        lower=0,
        upper=max(num_rows - grid_width, 0),
    )
    start_slide_idx = grid_width * start_row
    slides = islice(enumerate(state.deck.slides, start=1), start_slide_idx, None)

    rows = [Layout(name=str(r)) for r in range(grid_width)]
    cols = [[Layout(name=f"{r}-{c}") for c in range(grid_width)] for r, _ in enumerate(rows)]

    root.split_column(*rows)
    for row, layouts in zip(rows, cols):
        for layout in layouts:
            slide_number, slide = next(slides, (None, None))
            if slide is None:
                layout.update("")
            else:
                is_active_slide = slide is state.current_slide
                layout.update(
                    Panel(
                        slide.render(triggers=Triggers(times=(monotonic(),))),
                        title=joinify(" | ", [slide_number, slide.title]),
                        border_style=Style(
                            color="bright_cyan" if is_active_slide else None,
                            dim=not is_active_slide,
                        ),
                    )
                )
        row.split_row(*layouts)

    return root


def present_deck(state: State) -> None:
    rps_counter = RPSCounter()
    footer = Layout(Footer(state, rps_counter), name="footer", size=1)
    help = Layout(Help(state), name="help")

    def get_renderable() -> Layout:
        current_slide = state.deck[state.current_slide_idx]

        body = Layout(name="body", ratio=1)
        if state.mode is Mode.SLIDE:
            body.update(render_slide(state, current_slide))
        elif state.mode is Mode.DECK:
            split_layout_into_deck_grid(body, state)
        elif state.mode is Mode.HELP:
            body.update(help)
        elif state.mode is Mode.OPTIONS:
            body.update(state.options)
        else:  # pragma: unreachable
            raise UnknownModeError(f"Unrecognized mode: {state.mode!r}")

        root = Layout(name="root")
        root.split_column(body, footer)

        rps_counter.mark()

        return root

    with no_echo(), Live(
        get_renderable=get_renderable,
        console=state.console,
        screen=True,
        auto_refresh=True,
        refresh_per_second=TARGET_RPS,
        vertical_overflow="visible",
    ) as live:
        while True:
            handle_input(state, sys.stdin)
            live.refresh()



class SpielException(Exception):
    pass


class DuplicateInputHandler(SpielException):
    pass


class UnknownModeError(SpielException):
    pass


class NoDeckFound(SpielException):
    pass


class InvalidOptionValue(SpielException):
    pass



import sys
from dataclasses import dataclass, field
from pathlib import Path

from pendulum import DateTime, now
from rich.control import Control
from rich.style import Style
from rich.text import Text
from watchdog.events import FileSystemEvent, FileSystemEventHandler

from .load import load_deck_and_options
from .state import State


@dataclass
class DeckReloader(FileSystemEventHandler):
    state: State
    deck_path: Path
    last_reload: DateTime = field(default_factory=now)

    def on_modified(self, event: FileSystemEvent) -> None:
        self.last_reload = now()
        try:
            self.state.deck, _ = load_deck_and_options(self.deck_path)
            self.state.reset_trigger()
            self.state.set_message(
                lambda: Text(
                    f"Reloaded deck from {self.deck_path} {self.last_reload.diff_for_humans(None, False)}",
                    style=Style(color="bright_green"),
                )
            )
        except Exception:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            self.state.set_message(
                lambda: Text(
                    f"Error: {self.last_reload.diff_for_humans(None, False)}: {exc_obj!r}{Control.bell()}",
                    style=Style(color="bright_red"),
                )
            )

    def __hash__(self) -> int:
        return hash((type(self), id(self)))



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Callable, Union

from rich.console import ConsoleRenderable
from rich.text import Text

from .presentable import Presentable
from .triggers import Triggers

MakeRenderable = Callable[..., ConsoleRenderable]
RenderableLike = Union[MakeRenderable, ConsoleRenderable]


@dataclass
class Slide(Presentable):
    content: RenderableLike = field(default_factory=Text)

    def render(self, triggers: Triggers) -> ConsoleRenderable:
        if callable(self.content):
            return self.content(**self.get_render_kwargs(function=self.content, triggers=triggers))
        else:
            return self.content



from __future__ import annotations

import contextlib
import inspect
import string
import sys
import termios
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import dataclass
from enum import Enum, unique
from io import UnsupportedOperation
from itertools import product
from pathlib import Path
from textwrap import dedent
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    MutableMapping,
    NoReturn,
    Optional,
    TextIO,
    Tuple,
    Union,
)

import typer
from rich.control import Control
from rich.text import Text
from toml import TomlDecodeError
from typer import Exit

from .constants import EDITOR, PACKAGE_NAME
from .example import Example
from .exceptions import DuplicateInputHandler, InvalidOptionValue
from .modes import Mode
from .notebooks import NOTEBOOKS
from .options import Options
from .repls import REPLS
from .state import State

LFLAG = 3
CC = 6

try:
    ORIGINAL_TCGETATTR: Optional[List[Any]] = termios.tcgetattr(sys.stdin)
except (UnsupportedOperation, termios.error):
    ORIGINAL_TCGETATTR = None


@contextmanager
def no_echo() -> Iterator[None]:
    try:
        start_no_echo(sys.stdin)
        yield
    finally:
        reset_tty(sys.stdin)


def start_no_echo(stream: TextIO) -> None:
    if ORIGINAL_TCGETATTR is None:
        return

    mode = deepcopy(ORIGINAL_TCGETATTR)

    mode[LFLAG] = mode[LFLAG] & ~(termios.ECHO | termios.ICANON)
    mode[CC][termios.VMIN] = 1
    mode[CC][termios.VTIME] = 0

    termios.tcsetattr(stream.fileno(), termios.TCSADRAIN, mode)


def reset_tty(stream: TextIO) -> None:
    if ORIGINAL_TCGETATTR is None:
        return

    termios.tcsetattr(stream.fileno(), termios.TCSADRAIN, ORIGINAL_TCGETATTR)


@unique
class SpecialCharacters(Enum):
    Up = "↑"
    Down = "↓"
    Right = "→"
    Left = "←"
    CtrlUp = "ctrl-up"
    CtrlDown = "ctrl-down"
    CtrlRight = "ctrl-right"
    CtrlLeft = "ctrl-left"
    CtrlK = "ctrl-k"
    CtrlC = "ctrl-c"
    ShiftUp = "shift-up"
    ShiftDown = "shift-down"
    ShiftRight = "shift-right"
    ShiftLeft = "shift-left"
    CtrlShiftUp = "ctrl-shift-up"
    CtrlShiftDown = "ctrl-shift-down"
    CtrlShiftRight = "ctrl-shift-right"
    CtrlShiftLeft = "ctrl-shift-left"
    Backspace = "backspace"
    CtrlSpace = "ctrl-space"
    Enter = "enter"

    @classmethod
    def from_character(cls, character: str) -> SpecialCharacters:
        return SPECIAL_CHARACTERS[character]


SPECIAL_CHARACTERS = {
    "\x1b[A": SpecialCharacters.Up,
    "\x1b[B": SpecialCharacters.Down,
    "\x1b[C": SpecialCharacters.Right,
    "\x1b[D": SpecialCharacters.Left,
    "\x0b": SpecialCharacters.CtrlK,
    "\x1b[1;5A": SpecialCharacters.CtrlUp,
    "\x1b[1;5B": SpecialCharacters.CtrlDown,
    "\x1b[1;5C": SpecialCharacters.CtrlRight,
    "\x1b[1;5D": SpecialCharacters.CtrlLeft,
    "\x1b[1;2A": SpecialCharacters.ShiftUp,
    "\x1b[1;2B": SpecialCharacters.ShiftDown,
    "\x1b[1;2C": SpecialCharacters.ShiftRight,
    "\x1b[1;2D": SpecialCharacters.ShiftLeft,
    "\x1b[1;6A": SpecialCharacters.CtrlShiftUp,
    "\x1b[1;6B": SpecialCharacters.CtrlShiftDown,
    "\x1b[1;6C": SpecialCharacters.CtrlShiftRight,
    "\x1b[1;6D": SpecialCharacters.CtrlShiftLeft,
    "\x7f": SpecialCharacters.Backspace,
    "\x00": SpecialCharacters.CtrlSpace,
    "\n": SpecialCharacters.Enter,
}


def get_character(stream: TextIO) -> Union[str, SpecialCharacters]:
    result = stream.read(1)

    if result == "":  # this happens when stdin gets closed; equivalent to a quit
        raise Exit(code=0)

    if result[-1] == "\x1b":
        result += stream.read(2)

    if len(result) != 1 and result[-1] == "1":
        result += stream.read(3)

    try:
        return SpecialCharacters.from_character(result)
    except KeyError:
        return result


Character = Union[str, SpecialCharacters]
InputHandler = Callable[[State], Optional[NoReturn]]
InputHandlerKey = Tuple[Character, Mode]
InputHandlerDecorator = Callable[[InputHandler], InputHandler]
InputHandlers = MutableMapping[InputHandlerKey, InputHandler]

INPUT_HANDLERS: InputHandlers = {}  # type: ignore


@dataclass(frozen=True)
class InputHandlerHelpInfo:
    name: str
    help: str
    characters: Tuple[Character, ...]
    modes: List[Mode]


INPUT_HANDLER_HELP: List[InputHandlerHelpInfo] = []


def handle_input(
    state: State,
    stream: TextIO,
    handlers: InputHandlers = INPUT_HANDLERS,
) -> Optional[NoReturn]:
    character = get_character(stream)

    try:
        handler = handlers[(character, state.mode)]
    except KeyError:
        return None

    return handler(state)


def normalize_help(help: str) -> str:
    return dedent(help).replace("\n", " ").strip()


def input_handler(
    *characters: Character,
    modes: Optional[Iterable[Mode]] = None,
    handlers: InputHandlers = INPUT_HANDLERS,
    name: Optional[str] = None,
    help: str,
) -> InputHandlerDecorator:
    target_modes = list(modes or list(Mode))

    def registrar(func: InputHandler) -> InputHandler:
        for character, mode in product(characters, target_modes):
            key: InputHandlerKey = (character, mode)
            # Don't allow duplicate handlers to be registered inside this module,
            # but DO let end-users register them.
            if key in handlers and inspect.getmodule(func) == inspect.getmodule(input_handler):
                raise DuplicateInputHandler(
                    f"{character} is already registered as an input handler for mode {mode}"
                )
            handlers[key] = func

        INPUT_HANDLER_HELP.append(
            InputHandlerHelpInfo(
                name=name or " ".join(word.capitalize() for word in func.__name__.split("_")),
                help=normalize_help(help),
                characters=characters,
                modes=target_modes,
            )
        )

        return func

    return registrar


NOT_HELP = [Mode.SLIDE, Mode.DECK]


@input_handler(
    "h",
    help=f"Enter {Mode.HELP} mode.",
)
def help_mode(state: State) -> None:
    state.mode = Mode.HELP


@input_handler(
    "s",
    help=f"Enter {Mode.SLIDE} mode.",
)
def slide_mode(state: State) -> None:
    state.mode = Mode.SLIDE


@input_handler(
    "d",
    help=f"Enter {Mode.DECK} mode.",
)
def deck_mode(state: State) -> None:
    state.mode = Mode.DECK


@input_handler(
    "p",
    help=f"Enter {Mode.OPTIONS} mode.",
)
def options_mode(state: State) -> None:
    state.mode = Mode.OPTIONS


@input_handler(
    "e",
    modes=[Mode.OPTIONS],
    help=f"Open your $EDITOR ([bold]{EDITOR}[/bold]) to edit options (as TOML).",
)
def edit_options(state: State) -> None:
    with suspend_live(state):
        new_toml = state.options.as_toml()
        while True:
            new_toml = _clean_toml(typer.edit(text=new_toml, extension=".toml", require_save=False))
            try:
                state.options = Options.from_toml(new_toml)
                return
            except TomlDecodeError as e:
                new_toml = f"{new_toml}\n\n# Parse Error: {e}\n"
            except InvalidOptionValue as e:
                new_toml = f"{new_toml}\n\n# Invalid Option Value: {e}\n"
            except Exception as e:
                new_toml = f"{new_toml}\n\n# Error: {e}\n"


def _clean_toml(s: str) -> str:
    return "\n".join(line for line in s.splitlines() if (line and not line.startswith("#")))


@input_handler(
    SpecialCharacters.Right,
    "f",
    modes=NOT_HELP,
    help="Move to the next slide.",
)
def next_slide(state: State) -> None:
    state.next_slide()


@input_handler(
    SpecialCharacters.Left,
    "b",
    modes=NOT_HELP,
    help="Move to the previous slide.",
)
def previous_slide(state: State) -> None:
    state.previous_slide()


@input_handler(
    SpecialCharacters.Up,
    modes=[Mode.DECK],
    help="Move to the previous deck grid row.",
)
def up_grid_row(state: State) -> None:
    state.previous_slide(move=state.deck_grid_width)


@input_handler(
    SpecialCharacters.Down,
    modes=[Mode.DECK],
    help="Move to the next deck grid row.",
)
def down_grid_row(state: State) -> None:
    state.next_slide(move=state.deck_grid_width)


@input_handler(
    "j",
    modes=NOT_HELP,
    help="""\
    Press the action key, then a slide number (e.g., [bold]17[/bold]), then press [bold]enter[/bold], to jump to that slide.
    If the slide number is unambiguous, the jump will happen without needing to press [bold]enter[/bold]
    (e.g., you enter [bold]3[/bold] and there are only [bold]8[/bold] slides).
    """,
)
def jump_to_slide(state: State) -> None:
    slide_number = ""

    def display() -> None:
        state.set_message(Text(f"Jumping to slide {slide_number}..."))

    def jump() -> None:
        state.clear_message()
        if slide_number == "":
            return
        state.jump_to_slide(int(slide_number) - 1)
        return

    display()

    while True:
        char = get_character(sys.stdin)

        if char is SpecialCharacters.Backspace:
            slide_number = slide_number[:-1]
        elif char is SpecialCharacters.Enter:
            return jump()
        elif isinstance(char, SpecialCharacters):
            continue
        elif char in string.digits:
            slide_number += char

        display()

        if len(slide_number) == len(str(len(state.deck))):
            return jump()


@input_handler(
    "t",
    modes=[Mode.SLIDE],
    help="Trigger the slide: marks the current time and make it available to the slide's content rendering function.",
)
def trigger(state: State) -> None:
    state.trigger()


@input_handler(
    "r",
    modes=[Mode.SLIDE],
    help="Reset the trigger state to as if the slide just started being displayed.",
)
def reset_trigger(state: State) -> None:
    state.reset_trigger()


@contextlib.contextmanager
def suspend_live(state: State) -> Iterator[None]:
    live = state.console._live

    if live is None:
        yield
        return

    live.stop()
    yield
    live.start(refresh=True)


@input_handler(
    "e",
    modes=[Mode.SLIDE],
    help=f"Open your $EDITOR ([bold]{EDITOR}[/bold]) on the source of an [bold]Example[/bold] slide. If the current slide is not an [bold]Example[/bold], do nothing.",
)
def edit_example(state: State) -> None:
    example = state.current_slide
    if isinstance(example, Example):
        with suspend_live(state):
            example.source = typer.edit(
                text=example.source, extension=Path(example.name).suffix, require_save=False
            )
            example.clear_cache()


@input_handler(
    "i",
    name="Start REPL",
    modes=NOT_HELP,
    help=f"Start an [link=https://ipython.readthedocs.io/en/stable/overview.html]IPython REPL[/link].",
)
def open_repl(state: State) -> None:
    with suspend_live(state):
        reset_tty(sys.stdin)
        state.console.print(Control.clear())
        state.console.print(Control.move_to(0, 0))

        try:
            REPLS[state.options.repl]()
        finally:
            start_no_echo(sys.stdin)


@input_handler(
    "n",
    name="Open Notebook",
    modes=NOT_HELP,
    help=f"Open a Jupyter Notebook in your terminal using [link=https://github.com/davidbrochart/nbterm]nbterm[/link].",
)
def open_notebook(state: State) -> None:
    with suspend_live(state):
        reset_tty(sys.stdin)
        state.console.print(Control.clear())
        state.console.print(Control.move_to(0, 0))

        try:
            NOTEBOOKS[state.options.notebook](state)
        finally:
            start_no_echo(sys.stdin)


@input_handler(
    SpecialCharacters.CtrlK,
    SpecialCharacters.CtrlC,
    help=f"Exit {PACKAGE_NAME}.",
)
def exit(state: State) -> None:
    raise Exit(code=0)



import inspect
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, Mapping, Optional

from rich.console import ConsoleRenderable

from .triggers import Triggers


@dataclass
class Presentable:  # Why not an ABC? https://github.com/python/mypy/issues/5374
    title: str = ""
    notebook: Optional[Path] = None

    def render(self, triggers: Triggers) -> ConsoleRenderable:
        raise NotImplementedError

    def get_render_kwargs(self, function: Callable, triggers: Triggers) -> Mapping[str, Any]:
        signature = inspect.signature(function)

        kwargs: Dict[str, Any] = {}
        if "triggers" in signature.parameters:
            kwargs["triggers"] = triggers

        return kwargs



from __future__ import annotations

from dataclasses import dataclass
from math import floor
from pathlib import Path
from typing import Iterable, NamedTuple

from PIL import Image as Img
from rich.color import Color
from rich.console import Console, ConsoleOptions, JustifyMethod
from rich.segment import Segment
from rich.style import Style
from rich.text import Text

from .utils import chunks


class ImageSize(NamedTuple):
    width: int
    height: int


@dataclass(frozen=True)
class Image:
    img: Img
    justify: JustifyMethod = "center"

    @classmethod
    def from_file(cls, path: Path, justify: JustifyMethod = "center") -> Image:
        return cls(img=Img.open(path), justify=justify)

    def _determine_size(self, options: ConsoleOptions) -> ImageSize:
        width, height = self.img.size

        # multiply the max height by 2, because we're going to print 2 "pixels" per row
        max_height = options.height * 2 if options.height else None
        if max_height:
            width, height = width * max_height / self.img.height, max_height

        if width > options.max_width:
            width, height = options.max_width, height * options.max_width / width

        return ImageSize(floor(width), floor(height))

    def _resize(self, size: ImageSize) -> Img:
        return self.img.resize(
            size=size,
            resample=Img.LANCZOS,
        )

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> Iterable[Segment]:
        size = self._determine_size(options)
        resized = self._resize(size)

        rows = [
            [
                Text(
                    "▀",
                    Style(
                        color=Color.from_rgb(*top),
                        bgcolor=Color.from_rgb(*bottom) if bottom else None,
                    ),
                )
                # ... produce one row of text, using upper-half-blocks for the top image row and the background for the bottom image row
                for top, bottom in zip(top_row, bottom_row)
            ]
            # for each pair of rows in the image...
            for top_row, bottom_row in chunks(
                chunks(resized.getdata(), size.width), 2, fill_value=[None] * size.width
            )
        ]

        text = Text("\n", justify=self.justify).join(Text("").join(row) for row in rows)

        yield from text.__rich_console__(console, options)



from dataclasses import dataclass

from click._termui_impl import Editor  # type: ignore
from rich.align import Align
from rich.console import Console, ConsoleRenderable, RenderGroup
from rich.padding import Padding
from rich.style import Style
from rich.table import Column, Table
from rich.text import Text

from .constants import PACKAGE_NAME, __python_version__, __rich_version__, __version__
from .input import INPUT_HANDLER_HELP, SpecialCharacters
from .modes import Mode
from .state import State


@dataclass
class Help:
    state: State

    def __rich__(self) -> ConsoleRenderable:
        action_table = Table(
            Column(
                "Action",
                style=Style(bold=True),
            ),
            Column(
                "Keys",
                style=Style(bold=True),
                justify="center",
            ),
            Column(
                "Modes",
                justify="center",
            ),
            Column(
                "Description",
            ),
            show_lines=True,
        )

        for info in INPUT_HANDLER_HELP:
            action_table.add_row(
                Text(info.name),
                Text("  ").join(
                    Text(c.value if isinstance(c, SpecialCharacters) else c)
                    for c in info.characters
                ),
                Text(", ").join(Text(mode.value) for mode in info.modes)
                if len(info.modes) != len(list(Mode))
                else Text("any", style=Style(italic=True)),
                Text.from_markup(info.help),
            )

        return Padding(
            RenderGroup(
                Align.center(action_table),
                Align.center(version_details(self.state.console)),
            ),
            pad=(0, 1),
        )


def version_details(console: Console) -> ConsoleRenderable:
    table = Table(
        Column(justify="right"),
        Column(justify="left"),
        show_header=False,
        box=None,
    )

    table.add_row(f"{PACKAGE_NAME.capitalize()} Version", __version__)
    table.add_row("Rich Version", __rich_version__)
    table.add_row("Python Version", __python_version__, end_section=True)

    table.add_row(
        "Color System",
        Text(
            console.color_system or "unknown",
            style=Style(color="red" if console.color_system != "truecolor" else "green"),
        ),
    )
    table.add_row(
        "Console Dimensions",
        Text(f"{console.width} cells wide, {console.height} cells tall"),
        end_section=True,
    )

    editor = Editor().get_editor()
    table.add_row(
        "Editor",
        Text(editor),
        end_section=True,
    )

    return table



from collections import deque
from time import monotonic
from typing import Deque, Optional

from .constants import TARGET_RPS


class RPSCounter:
    def __init__(self, render_history_length: Optional[int] = None) -> None:
        if render_history_length is None:
            render_history_length = 3 * TARGET_RPS

        self.render_time_history: Deque[float] = deque(maxlen=render_history_length)

    def mark(self) -> None:
        self.render_time_history.append(monotonic())

    def renders_per_second(self) -> float:
        num_samples = len(self.render_time_history)
        if num_samples < 2:
            return 0

        return num_samples / (self.render_time_history[-1] - self.render_time_history[0])



from __future__ import annotations

from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from tempfile import TemporaryDirectory
from time import monotonic
from types import TracebackType
from typing import Callable, List, Optional, Type, Union

from rich.console import Console
from rich.style import Style
from rich.text import Text

from .constants import PACKAGE_NAME
from .deck import Deck
from .load import load_deck_and_options
from .modes import Mode
from .options import Options
from .presentable import Presentable

TextLike = Union[Text, Callable[[], Text]]


@dataclass
class State:
    console: Console
    deck: Deck
    options: Options
    _current_slide_idx: int = 0
    _mode: Mode = Mode.SLIDE
    _message: TextLike = Text("")
    trigger_times: List[float] = field(default_factory=list)

    @classmethod
    def from_file(cls, path: Path) -> State:
        deck, options = load_deck_and_options(path)
        return cls(console=Console(), deck=deck, options=options)

    @property
    def mode(self) -> Mode:
        return self._mode

    @mode.setter
    def mode(self, mode: Mode) -> None:
        self._mode = mode
        self.reset_trigger()

    @property
    def current_slide_idx(self) -> int:
        return self._current_slide_idx

    @current_slide_idx.setter
    def current_slide_idx(self, idx: int) -> None:
        self._current_slide_idx = max(0, min(len(self.deck) - 1, idx))
        self.reset_trigger()

    def next_slide(self, move: int = 1) -> None:
        if self.current_slide_idx == len(self.deck) - 1:
            return
        self.current_slide_idx += move

    def previous_slide(self, move: int = 1) -> None:
        if self.current_slide_idx == 0:
            return
        self.current_slide_idx -= move

    def jump_to_slide(self, idx: int) -> None:
        self.current_slide_idx = idx

    @property
    def current_slide(self) -> Presentable:
        return self.deck[self.current_slide_idx]

    @property
    def message(self) -> Text:
        if callable(self._message):
            try:
                return self._message()
            except Exception:
                return Text(
                    "Internal Error: failed to display message.",
                    style=Style(color="bright_red"),
                )
        else:
            return self._message

    def set_message(self, message: TextLike) -> None:
        self._message = message

    def clear_message(self) -> None:
        self.set_message(Text(""))

    @property
    def deck_grid_width(self) -> int:
        return max(self.console.size.width // 30, 1)

    def trigger(self) -> None:
        self.trigger_times.append(monotonic())

    def reset_trigger(self) -> None:
        self.trigger_times.clear()
        self.trigger()

    @cached_property
    def _tmp_dir(self) -> TemporaryDirectory:
        return TemporaryDirectory(prefix=f"{PACKAGE_NAME}-")

    @cached_property
    def tmp_dir(self) -> Path:
        return Path(self._tmp_dir.name)

    def __enter__(self) -> State:
        return self

    def __exit__(
        self,
        exctype: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ) -> None:
        self._tmp_dir.cleanup()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['spiel', 'spiel.demo']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.2.0,<9.0.0',
 'ipykernel>=5.5.3,<6.0.0',
 'ipython>=7.23.0,<8.0.0',
 'nbterm>=0.0.7,<0.0.8',
 'numpy>=1.20.3,<2.0.0',
 'pendulum>=2.1.2,<3.0.0',
 'rich>=10.0.0,<11.0.0',
 'toml>=0.10.2,<0.11.0',
 'typer>=0.3.2,<0.4.0',
 'uniplot>=0.4.4,<0.5.0',
 'watchdog>=2.0.2,<3.0.0']

entry_points = \
{'console_scripts': ['spiel = spiel.main:app']}

setup_kwargs = {
    'name': 'spiel',
    'version': '0.2.0',
    'description': 'A framework for building and presenting richly-styled presentations in your terminal using Python.',
    'long_description': "# Spiel\n\n[![PyPI](https://img.shields.io/pypi/v/spiel)](https://pypi.org/project/spiel/)\n[![Documentation Status](https://readthedocs.org/projects/spiel/badge/?version=latest)](https://spiel.readthedocs.io/en/latest/?badge=latest)\n[![PyPI - License](https://img.shields.io/pypi/l/spiel)](https://pypi.org/project/spiel/)\n\n[![pre-commit.ci status](https://results.pre-commit.ci/badge/github/JoshKarpel/spiel/main.svg)](https://results.pre-commit.ci/latest/github/JoshKarpel/spiel/main)\n[![codecov](https://codecov.io/gh/JoshKarpel/spiel/branch/main/graph/badge.svg?token=2sjP4V0AfY)](https://codecov.io/gh/JoshKarpel/spiel)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n[![GitHub issues](https://img.shields.io/github/issues/JoshKarpel/spiel)](https://github.com/JoshKarpel/spiel/issues)\n[![GitHub pull requests](https://img.shields.io/github/issues-pr/JoshKarpel/spiel)](https://github.com/JoshKarpel/spiel/pulls)\n\nSpiel is a framework for building and presenting richly-styled presentations in your terminal using Python.\n\nTo see what Spiel can do without installing it, you can view the demonstration deck in a container:\n```bash\n$ docker run -it --rm ghcr.io/joshkarpel/spiel\n```\nAlternatively, install Spiel (`pip install spiel`) and run this command to view the demonstration deck:\n```bash\n$ spiel demo present\n```\n\n## Sandboxed Execution via Containers\n\nSpiel presentations are live Python code: they can do anything that Python can do.\nYou may want to run untrusted presentations (or even your own presentations) inside a container (but remember, even containers are not perfectly safe!).\nWe produce a [container image](https://github.com/users/JoshKarpel/packages/container/package/spiel)\nthat can be run by (for example) Docker.\n\nPresentations without extra Python dependencies might just need to be bind-mounted into the container.\nFor example, if your demo file is at `$PWD/presentation/deck.py`, you could do\n```bash\n$ docker run -it --rm --mount type=bind,source=$PWD/presentation,target=/presentation ghcr.io/joshkarpel/spiel spiel present /presentation/deck.py\n```\n\nIf the presentation has extra dependencies (like other Python packages),\nwe recommend building a new image that inherits our image (e.g., `FROM ghcr.io/joshkarpel/spiel:vX.Y.Z`).\nSpiel's image itself inherits from the [Python base image](https://hub.docker.com/_/python).\n\n## Supported Systems\n\nSpiel currently relies on underlying terminal mechanisms that are only available on POSIX systems (e.g., Linux and MacOS).\nIf you're on Windows, you can use the [Windows Subsystem for Linux](https://docs.microsoft.com/en-us/windows/wsl/) to run Spiel.\n",
    'author': 'JoshKarpel',
    'author_email': 'josh.karpel@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/JoshKarpel/spiel',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from dataclasses import asdict, dataclass, fields
from pathlib import Path
from typing import Any, Mapping

import toml
from rich.align import Align
from rich.console import ConsoleRenderable
from rich.padding import Padding
from rich.table import Column, Table

from .constants import PACKAGE_NAME
from .exceptions import InvalidOptionValue
from .notebooks import NOTEBOOKS
from .repls import REPLS


@dataclass
class Options:
    repl: str = "ipython"
    notebook: str = "nbterm"
    footer_time_format: str = "YYYY-MM-DD hh:mm A"
    profiling: bool = False

    def __post_init__(self) -> None:
        if self.repl not in REPLS:
            raise InvalidOptionValue(f"repl must be one of: {set(REPLS.keys())}")

        if self.notebook not in NOTEBOOKS:
            raise InvalidOptionValue(f"notebook must be one of: {set(NOTEBOOKS.keys())}")

    def as_dict(self) -> Mapping[str, Any]:
        return asdict(self)

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> "Options":
        fields_by_name = {field.name: field for field in fields(cls)}
        only_valid = {k: fields_by_name[k].type(v) for k, v in d.items() if k in fields_by_name}
        return cls(**only_valid)

    def as_toml(self) -> str:
        return toml.dumps({PACKAGE_NAME: self.as_dict()})

    @classmethod
    def from_toml(cls, t: str) -> "Options":
        return cls.from_dict(toml.loads(t).get(PACKAGE_NAME, {}))

    def save(self, path: Path) -> Path:
        path.write_text(self.as_toml())
        return path

    @classmethod
    def load(cls, path: Path) -> "Options":
        return cls.from_toml(path.read_text())

    def __rich__(self) -> ConsoleRenderable:
        table = Table(
            Column("Option"),
            Column("Type", justify="center"),
            Column("Value"),
            show_lines=True,
        )

        fields_by_name = {field.name: field for field in fields(self)}

        for key, value in self.as_dict().items():
            table.add_row(key, fields_by_name[key].type.__name__, str(value))

        return Padding(
            Align.center(table),
            pad=(0, 1),
        )



from dataclasses import dataclass

from pendulum import now
from rich.console import ConsoleRenderable
from rich.style import Style
from rich.table import Column, Table
from rich.text import Text

from .modes import Mode
from .rps import RPSCounter
from .state import State
from .utils import drop_nones, joinify


@dataclass
class Footer:
    state: State
    rps_counter: RPSCounter

    @property
    def longest_slide_number_length(self) -> int:
        num_slides = len(self.state.deck)
        return len(str(num_slides))

    def __rich__(self) -> ConsoleRenderable:
        grid = Table.grid(
            *drop_nones(
                Column(
                    style=Style(dim=True),
                    justify="left",
                ),
                Column(
                    style=Style(bold=True),
                    justify="center",
                ),
                Column(
                    style=Style(dim=True),
                    justify="right",
                )
                if self.state.options.profiling
                else None,
                Column(
                    style=Style(dim=True),
                    justify="right",
                ),
                Column(
                    style=Style(dim=True),
                    justify="right",
                ),
            ),
            expand=True,
            padding=1,
        )
        grid.add_row(
            *drop_nones(
                Text(
                    joinify(
                        " | ",
                        [
                            self.state.deck.name,
                            self.state.current_slide.title
                            if self.state.mode is Mode.SLIDE
                            else None,
                        ],
                    )
                ),
                self.state.message,
                Text(f"{self.rps_counter.renders_per_second() :.2f} RPS")
                if self.state.options.profiling
                else None,
                now().format(self.state.options.footer_time_format),
                Text(
                    f"[{self.state.current_slide_idx + 1:>0{self.longest_slide_number_length}d} / {len(self.state.deck)}]"
                )
                if self.state.mode is not Mode.HELP
                else Text(Mode.HELP.value, style=Style(italic=True)),
            )
        )
        return grid



from __future__ import annotations

import shlex
import sys
import tempfile
from dataclasses import dataclass
from pathlib import Path
from subprocess import PIPE, STDOUT, run
from typing import Callable, Optional, Sequence

from rich.align import Align
from rich.console import ConsoleRenderable
from rich.layout import Layout
from rich.panel import Panel
from rich.syntax import Syntax
from rich.text import Text

from .presentable import Presentable
from .triggers import Triggers


@dataclass
class CachedExample:
    trigger_number: int
    input: str
    output: Optional[str]


def example_panels(example: Example) -> ConsoleRenderable:
    root = Layout()
    root.split_column(
        Layout(
            Align.center(
                Panel(
                    example.input,
                    title=example.name,
                    title_align="left",
                    expand=False,
                )
            )
        ),
        Layout(
            Align.center(
                Panel(
                    example.output,
                    title=f"$ {example.display_command}",
                    title_align="left",
                    expand=False,
                )
                if example.output is not None
                else Text(" ")
            )
        ),
    )
    return root


ExampleLayout = Callable[["Example"], ConsoleRenderable]


@dataclass
class Example(Presentable):
    source: str = ""
    command: Sequence[str] = (sys.executable,)
    name: str = "example.py"
    language: str = "python"
    _layout: ExampleLayout = example_panels
    _cache: Optional[CachedExample] = None

    def layout(self, function: ExampleLayout) -> ExampleLayout:
        self._layout = function  # type: ignore
        return function

    @property
    def display_command(self) -> str:
        return shlex.join([Path(self.command[0]).stem, *self.command[1:], self.name])

    def execute(self) -> str:
        with tempfile.TemporaryDirectory() as tmpdir:
            dir = Path(tmpdir)
            file = dir / self.name
            file.write_text(self.source)
            result = run([*self.command, file], stdout=PIPE, stderr=STDOUT, text=True)
        return result.stdout

    @property
    def input(self) -> Syntax:
        input = (self._cache.input or "") if self._cache is not None else ""
        return Syntax(
            input,
            lexer_name=self.language,
            code_width=max(len(line) for line in input.splitlines()),
        )

    @property
    def output(self) -> Optional[Text]:
        return (
            Text(self._cache.output)
            if (self._cache is not None and self._cache.output is not None)
            else None
        )

    def clear_cache(self) -> None:
        self._cache = None

    def render(self, triggers: Triggers) -> ConsoleRenderable:
        if self._cache is None:
            self._cache = CachedExample(len(triggers), self.source, None)
        elif self._cache.trigger_number != len(triggers):
            self._cache = CachedExample(len(triggers), self.source, self.execute())

        return self._layout(self, **self.get_render_kwargs(function=self._layout, triggers=triggers))  # type: ignore
